@@ -172,10 +172,10 @@ class _Project(_Base):
 
 
 class _Datasource(_Base):
-    def __init__(self, token, host, key, ws):
-        super().__init__(token, host)
+    def __init__(self, dfl, key):
+        super().__init__(dfl.token, dfl.host)
+        self.dfl = dfl
         self.key = key
-        self.ws = ws
         self.trainers = {}
 
     def add_trainer(self, key, train_callback, test_callback, description=None, model_path=None):
@@ -191,6 +191,13 @@ class _Datasource(_Base):
         if model_path is not None:
             self.trainers[key]['model_path'] = model_path
 
+        for p in self.dfl.project_participants:
+            project_key = p['project_key']
+            datasource_key = p['datasource_key']
+            info = self._make_request('GET', f'/projects/{project_key}')
+            threads_key = get_threads_key(project_key, datasource_key)
+            self.dfl.on_round_threads[threads_key] = gevent.spawn(self.dfl.train_and_test_callback, datasource_key, info)
+
 
 class DynamoFL(_Base):
     def __init__(self, token, host='https://api.dynamofl.com', metadata=None):
@@ -243,7 +250,7 @@ class DynamoFL(_Base):
 
             info = self._make_request('GET', f'/projects/{project_key}')
             threads_key = get_threads_key(project_key, datasource_key)
-            self.on_round_threads[threads_key] = gevent.spawn(self.train_callback, datasource_key, info)
+            self.on_round_threads[threads_key] = gevent.spawn(self.train_and_test_callback, datasource_key, info)
 
 
         if 'data' in j and 'project' in j['data'] and 'key' in j['data']['project']:
@@ -253,11 +260,13 @@ class DynamoFL(_Base):
                 threads_key = get_threads_key(project_key, p['datasource_key'])
                 if threads_key in self.on_round_threads:
                     self.on_round_threads[threads_key].kill()
+                if project_key == p['project_key']:
+                    self.on_round_threads[threads_key] = gevent.spawn(self.train_and_test_and_end_callback, p['datasource_key'], j['data']['project'])
+
                     # del self.projects[project_key]
-                    self.project_participants = list(filter(lambda x : x['project_key'] != project_key, self.project_participants))
-                    if not len(self.project_participants):
-                        self.ws.close()
-                        return
+            # if not len(self.project_participants):
+            #     self.ws.close()
+            #     return
 
 
         elif j['event'] == 'round-complete':
@@ -266,7 +275,7 @@ class DynamoFL(_Base):
                 if threads_key in self.on_round_threads:
                     self.on_round_threads[threads_key].kill()
                 if project_key == p['project_key']:
-                    self.on_round_threads[threads_key] = gevent.spawn(self.train_callback, p['datasource_key'], j['data']['project'])
+                    self.on_round_threads[threads_key] = gevent.spawn(self.train_and_test_callback, p['datasource_key'], j['data']['project'])
 
 
         elif j['event'] == 'round-error':
@@ -282,7 +291,7 @@ class DynamoFL(_Base):
         print('Connection error:')
         print(error)
 
-    def train_callback(self, datasource_key, project_info):
+    def train_and_test_callback(self, datasource_key, project_info):
         project = _Project(self.token, self.host, project_info['key'], self.ws)
 
         # on some project round completed
@@ -298,16 +307,15 @@ class DynamoFL(_Base):
             model_path = self.datasources[datasource_key].trainers[trainer_key]['model_path']
 
         current_round = project_info['currentRound']
+        if project_info['isComplete']:
+            current_round += 1
         ext = project_info['modelType']
         prev_round = current_round - 1
         federated_model_path = get_federated_path(model_path, ext, datasource_key, prev_round)
 
         # Pull
-        if os.path.isfile(federated_model_path):
-            print(f'>>> Already downloaded round ({prev_round}) federated model.')
-        else:
-            print(f'>>> Waiting to download round ({prev_round}) federated model...')
-            project.pull_model(federated_model_path, round=prev_round)
+        print(f'>>> Waiting to download round ({prev_round}) federated model...')
+        project.pull_model(federated_model_path, round=prev_round)
 
         # Test
         if test == None:
@@ -324,15 +332,21 @@ class DynamoFL(_Base):
             print()
 
         # Push
-        new_model_path = get_trained_path(model_path, ext, datasource_key, current_round)
+        if not project_info['isComplete']:
+            new_model_path = get_trained_path(model_path, ext, datasource_key, current_round)
 
-        print('>>> Training weights on local model...')
-        train(datasource_key, federated_model_path, new_model_path, project_info)
+            print('>>> Training weights on local model...')
+            train(datasource_key, federated_model_path, new_model_path, project_info)
+
+            print(f'>>> Uploading round ({current_round}) trained model...')
+            project.push_model(new_model_path, datasource_key)
+            print('Done.')
+            print()
+
+    def train_and_test_and_end_callback(self, datasource_key, project_info):
+        self.train_and_test_callback(datasource_key, project_info)
+        self.project_participants = list(filter(lambda x : x['project_key'] != project_info['key'], self.project_participants))
 
-        print(f'>>> Uploading round ({current_round}) trained model...')
-        project.push_model(new_model_path, datasource_key)
-        print('Done.')
-        print()
 
     # creates a new datasource in the api
     def attach_datasource(self, key, name=None, metadata=None):
@@ -363,10 +377,13 @@ class DynamoFL(_Base):
                 'trainer_key': i['trainerKey']
             })
 
-        ds = _Datasource(self.token, self.host, key, self.ws)
+        ds = _Datasource(self, key)
         self.datasources[key] = ds
         return ds
 
+    def delete_datasource(self, key):
+        return self._make_request('DELETE', f'/datasources/{key}')
+
     def get_user(self):
         return self._make_request('GET', '/user')
 
